(function()
{
	"use strict";
	
    window.resources = new Resources();
	var MODE_NONE = -1;
	var MODE_GETFILESIZES = 0;
	var MODE_LOAD = 1;
	var MODE_MAINMENU = 2;
	var MODE_PLAY = 3;
	var MODE_DONE = 4;
    var MODE_LEVELMENU = 5;
	var MODETRANSITIONTIME = 1;
    
    window.BUTTON_ALWAYS = 0;
    window.BUTTON_RANDOM = 1;
    window.BUTTON_STORY = 2;
	
	window.mouseOverLevelNumber = null;
	var mode = MODE_GETFILESIZES;
	window.timeModeStarted = 0;
	window.timeModeWillEnd = Number.POSITIVE_INFINITY;
	var nextModeFunction = null;
	window.theGameState = null;
	var lastMouseX = null;
	var lastMouseY = null;
	window.levelPlayerIsUpTo = 0;
    window.maxLevelReached = 0;
	window.infLevelsSolved = 0;
	window.frameRate = null;
	var frameCount = 0;
	var lastFrameTime = null;
	
	var thePlaylist = ["ld31a", "ld31b", "ld31c", "ld31d"];
	
    window.callCmgFunction = function()
    {
        if(parent && parent.cmgGameEvent)
        {
            parent.cmgGameEvent.apply(this, arguments);
        }
        else
        {
            var msg = "cmgGameEvent called with arguments: " + JSON.stringify(arguments);
            console.log(msg)
        }
    }
    
    function validStartup()
    {
        var re = /^(.*\.)?coolmath-games\.com$/;
        return parent && parent.document.location.host.match(re) !== null;
    }
    
    window.unlockAllLevels = function()
    {
        window.maxLevelReached = storyParams.length - 1;
        storeLevel();
    }
    
	window.onload = function()
	{
        //if(!validStartup()) return;
		// check requirements
		var missingFeatures = [];
		if(!XMLHttpRequest) {missingFeatures.push("XMLHttpRequest");}
		if(!document.createElement('canvas')) {missingFeatures.push("canvas");}
		if(missingFeatures.length > 0)
		{
			var message = "Your browser does not support features required by this game: ";
			for(var i = 0; i < missingFeatures.length; i++)
			{
				if(i > 0) message += ', ';
				message += missingFeatures[i];
			}
			message += ". Apologies - try with a different browser.";
			document.body.innerHTML = message;
			return;
		}
		
		// check if the level the player is up to is in localStorage
		var lev;
		if(localStorage)
		{
			var storedLevel = localStorage.getItem('disc_levelPlayerIsUpTo');
			if(storedLevel !== null)
			{
				lev = parseInt(storedLevel);
				if(lev >= storyParams.length) lev = 0;
				window.levelPlayerIsUpTo = lev;
			}
            
            var storedMaxLevel = localStorage.getItem('disc_maxLevelReached');
            if(storedMaxLevel !== null)
            {
                window.maxLevelReached = parseInt(storedMaxLevel);
            }
			
			var storedInfLevels = localStorage.getItem('disc_infLevelsSolved');
			if(storedInfLevels !== null)
			{
				window.infLevelsSolved = parseInt(storedInfLevels);
			}
		}
		
		// start the animation loop
		requestAnimationFrame(frame);
		// request resources, and set a callback for when we know the file sizes
		resources.reqImage("title");
		
		for(var plidx = 0; plidx < thePlaylist.length; plidx++)
		{
			resources.reqAudio(thePlaylist[plidx]);
		}
        resources.reqAudio("place");
		resources.getAllFileSizes(handleGotFileSizes);
		// setup listeners
		gameCanvas.addEventListener("mousemove", handleMouseMove);
        gameCanvas.addEventListener('dblclick', handleDoubleClick);
		gameCanvas.addEventListener("click", handleClick);
	}
	
	function handleGotFileSizes()
	{
		// start downloading the files, and set a callback for when they arrive
		resources.downloadAll(handleDownloadedFiles);
		mode = MODE_LOAD;
	}
	
	function handleDownloadedFiles()
	{
		// all the resources are downloaded! we can start!
		mode = MODE_MAINMENU;
		window.timeModeStarted = timeSecs();
		window.timeModeWillEnd = Number.POSITIVE_INFINITY;
	}
	
	window.timeSecs = function()
	{
		return (new Date()).getTime() / 1000;
	}
	
	function chooseTimeModeWillEnd()
	{
		var t = timeSecs();
		var tSince = t - timeModeStarted;
		var step = Math.min(tSince, MODETRANSITIONTIME);
		return Math.min(t + step, timeModeWillEnd);
	}
	
	function startMusicIfNotStarted()
	{
		var playlist = [];
		for(var i = 0; i < thePlaylist.length; i++)
		{
			playlist.push(resources.data[thePlaylist[i]]);
		}
		if(!resources.musicStarted) resources.startMusic(playlist, 0);
	}
	
	window.setStoryMode = function()
	{
		window.timeModeWillEnd = chooseTimeModeWillEnd();
		nextModeFunction = startStoryModeNew;
	};
	
	window.continueStory = function()
	{
		window.timeModeWillEnd = chooseTimeModeWillEnd();
		nextModeFunction = startStoryModeContinue;
	};
	
	function startStoryModeNew()
	{
        callCmgFunction("start");
		mode = MODE_PLAY;
		levelPlayerIsUpTo = 0;
		storeLevel();
		storyModePlayLevel(0);
		window.timeModeStarted = timeSecs();
		startMusicIfNotStarted();
	}
	
	function startStoryModeContinue()
	{
        callCmgFunction("start");
		mode = MODE_PLAY;
		storyModePlayLevel(levelPlayerIsUpTo);
		window.timeModeStarted = timeSecs();
		startMusicIfNotStarted();
	}
	
	window.setInfiniteMode = function()
	{
		window.timeModeWillEnd = chooseTimeModeWillEnd();
		nextModeFunction = startInfiniteMode;
	};
	
	window.storeLevel = function()
	{
		localStorage.setItem('disc_levelPlayerIsUpTo', levelPlayerIsUpTo);
        localStorage.setItem('disc_maxLevelReached', maxLevelReached);
		localStorage.setItem('disc_infLevelsSolved', infLevelsSolved);
	}
	
	function startInfiniteMode()
	{
		mode = MODE_PLAY;
		infiniteModeGenerate();
		window.timeModeStarted = timeSecs();
		startMusicIfNotStarted();
	}
	
	window.gameOver = function()
	{
		window.timeModeWillEnd = chooseTimeModeWillEnd();
		nextModeFunction = handleGameOver;
	}
	
	function handleGameOver()
	{
		mode = MODE_DONE;
		levelPlayerIsUpTo = 0;
		storeLevel();
		window.timeModeStarted = timeSecs();
	}
	
	var mainMenuButtonSpecsNoContinue = {
		'storyMode': ["Story Mode", 0, setStoryMode, false],
		'infinite': ["Infinite Play", 2, setInfiniteMode, false]
	}
	
	var mainMenuButtonSpecsWithContinue = {
		'continueStory': ["Continue Story", 0, continueStory, false],
		'restartStory': ["Restart Story", 1, setStoryMode, false],
		'infinite': ["Infinite Play", 2, setInfiniteMode, false]
	}
	
	function quitToMenu()
	{
		timeModeWillEnd = chooseTimeModeWillEnd();
		nextModeFunction = handleQuitToMenu;
	}
	
	function handleQuitToMenu()
	{
		mode = MODE_MAINMENU;
		timeModeStarted = timeSecs();
		timeModeWillEnd = Number.POSITIVE_INFINITY;
		resources.haltSpeechSequence();
	}
    
    function jumpToLevelMenu()
    {
        timeModeWillEnd = chooseTimeModeWillEnd();
        nextModeFunction = handleJumpToLevelMenu;
    }
    
    function handleJumpToLevelMenu()
    {
        mode = MODE_LEVELMENU;
        timeModeStarted = timeSecs();
		timeModeWillEnd = Number.POSITIVE_INFINITY;
		resources.haltSpeechSequence();
    }
    
    function selectLevel(idx)
    {
        timeModeWillEnd = chooseTimeModeWillEnd();
        nextModeFunction = handleSelectLevel.bind(null, idx);
    }
    
    function handleSelectLevel(idx)
    {
		mode = MODE_PLAY;
        levelPlayerIsUpTo = idx;
        storeLevel();
		storyModePlayLevel(idx);
		window.timeModeStarted = timeSecs();
		startMusicIfNotStarted();
    }
	
	function regenerate()
	{
		if(theGameState.endedTime === null)
		{
			theGameState.endedTime = timeSecs();
		}
	}
	
	function resetEdges()
	{
		if(theGameState !== null) theGameState.selEdges = [];
	}
    
	var inGameButtons = [
		['Quit to menu', quitToMenu, false, BUTTON_ALWAYS],
		['Regenerate', regenerate, false, BUTTON_RANDOM /* only for random levels */],
		['Reset edges', resetEdges, false, BUTTON_ALWAYS],
        ['Level menu', jumpToLevelMenu, false, BUTTON_STORY /* only for story mode */]
	];
    
    window.canSelectLevel = function(lev)
    {
        return lev <= maxLevelReached;
    }
	
    function handleMouseLevelMenu(canv, lastMouseX, lastMouseY, isClick)
    {
		var mes = drawingMeasures(canv);
		var cpos = getTotalCameraPos();
        var nLevs = storyParams.length;
        var lev = null;
        
        var drawOffX = -2;
        var drawOffY = -1;
        
        for(var i = 0; i < nLevs; i++)
        {
            var cellx = (i % 5);
            var celly = Math.floor(i / 5);
            var drawx = mes.cx + (cellx + drawOffX + 0.5 - cpos[0]) * mes.scale;
            var drawy = mes.cy + (celly + drawOffY + 0.5 - cpos[1]) * mes.scale;
            var r = 0.4 * mes.scale;
            var diffx = lastMouseX - drawx;
            var diffy = lastMouseY - drawy;
            var distSq = diffx * diffx + diffy * diffy;
            if(distSq <= r * r)
            {
                lev = i;
            }
        }
        window.mouseOverLevelNumber = lev;
        
        if(isClick)
        {
            if(lev !== null && canSelectLevel(lev))
            {
                selectLevel(lev);
            }
        }
    }
	
	function frame()
	{
		// make sure the canvas size is set correctly
		gameCanvas.width = window.innerWidth;
		gameCanvas.height = window.innerHeight;
		var ctx = gameCanvas.getContext('2d');
		ctx.clearRect(0, 0, gameCanvas.width, gameCanvas.height);
		// get a time value
		var t = timeSecs();
		// update the camera position
		updateCameraPos(t);
		// draw the background grid
		drawGrid(gameCanvas, t);
		// draw based on mode
		if(mode == MODE_GETFILESIZES)
		{
			drawCentreText(gameCanvas, 'Loading...');
		}
		else if(mode == MODE_LOAD)
		{
			var loadString;
			if(resources.numDownloaded === resources.numRequests)
			{
				if(resources.useWebAudio)
				{
					loadString = "Decoding (this could take a minute)....";
				}
				else
				{
					loadString = "Decoding...";
				}
			}
			else
			{
				loadString = "Loading... (";
				var numer = (resources.totalLoaded / 1e6).toFixed(1);
				var denom = (resources.totalToLoad / 1e6).toFixed(1);
				loadString += numer + "/" + denom + "MB)";
			}
			drawCentreText(gameCanvas, loadString);
		}
		else if(mode == MODE_MAINMENU)
		{
			if(levelPlayerIsUpTo === 0)
			{
				drawMainMenu(gameCanvas, mainMenuButtonSpecsNoContinue, t);
			}
			else
			{
				drawMainMenu(gameCanvas, mainMenuButtonSpecsWithContinue, t);
			}
		}
		else if(mode == MODE_PLAY)
		{
			drawPlayMode(gameCanvas, theGameState, inGameButtons, t);
			if(lastMouseX !== null && lastMouseY !== null)
			{
				window.handleMousePlay(gameCanvas, lastMouseX, lastMouseY, inGameButtons, false)
			}
		}
		else if(mode == MODE_DONE)
		{
			drawDoneMode(gameCanvas, t);
		}
        else if(mode == MODE_LEVELMENU)
        {
            drawLevelMenu(gameCanvas, t);
			if(lastMouseX !== null && lastMouseY !== null)
			{
				handleMouseLevelMenu(gameCanvas, lastMouseX, lastMouseY, false)
			}
        }
		drawSubtitle(gameCanvas, mode == MODE_PLAY, theGameState, t);
		// check if the mode needs to change at the end of a transition
		if(t > timeModeWillEnd)
		{
			nextModeFunction();
			timeModeWillEnd = Number.POSITIVE_INFINITY;
		}
		
		// update framerate
		if(frameCount % 100 === 0)
		{
			if(lastFrameTime !== null)
			{
				frameRate = 100 / (t - lastFrameTime);
			}
			lastFrameTime = t;
		}
		frameCount++;
		
		// ask for another frame
		requestAnimationFrame(frame);
	}
	
	function getCoordsOfMouseEvent(evt)
	{
		// get mouse coordinates in canvas coords
        var rect = gameCanvas.getBoundingClientRect();
        var x = evt.clientX - rect.left;
        var y = evt.clientY - rect.top;
		
		return [x, y];
	}
	
	function handleMouseMove(evt)
	{
		var mCoords = getCoordsOfMouseEvent(evt);
		lastMouseX = mCoords[0];
		lastMouseY = mCoords[1];
		if(mode == MODE_MAINMENU)
		{
			if(levelPlayerIsUpTo === 0)
			{
				handleMouseMainMenu(gameCanvas, mCoords[0], mCoords[1], false, mainMenuButtonSpecsNoContinue);
			}
			else
			{
				handleMouseMainMenu(gameCanvas, mCoords[0], mCoords[1], false, mainMenuButtonSpecsWithContinue);
			}
		}
		else if(mode == MODE_PLAY)
		{
			handleMousePlay(gameCanvas, mCoords[0], mCoords[1], inGameButtons, false);
		}
	}
	
    function handleDoubleClick(e)
    {
        // don't let event bubble up
        e.preventDefault();
        e.stopPropagation();
        return false;
    }
    
	function handleClick(evt)
	{
		var mCoords = getCoordsOfMouseEvent(evt);
		
		if(mode == MODE_MAINMENU)
		{
			if(levelPlayerIsUpTo === 0)
			{
				handleMouseMainMenu(gameCanvas, mCoords[0], mCoords[1], true, mainMenuButtonSpecsNoContinue);
			}
			else
			{
				handleMouseMainMenu(gameCanvas, mCoords[0], mCoords[1], true, mainMenuButtonSpecsWithContinue);
			}
		}
		else if(mode == MODE_PLAY)
		{
			handleMousePlay(gameCanvas, mCoords[0], mCoords[1], inGameButtons, true);
		}
		else if(mode == MODE_DONE && resources.speechSequence === null)
		{
			// return to main menu
			mode = MODE_MAINMENU;
		}
        else if(mode == MODE_LEVELMENU)
        {
            handleMouseLevelMenu(gameCanvas, lastMouseX, lastMouseY, true)
        }
	}
})();


