(function()
{
	"use strict";
	
	// a structure containing offsets and speeds for blinking cells
	
	// just plucked a number that hopefully won't make it look repetitive
	var BLINK_MODULUS = 227;
	var blinkSpeeds = [];
	var blinkPhases = [];
	(function(){
		for(var i = 0; i < BLINK_MODULUS; i++)
		{
			blinkSpeeds.push(Math.random());
			blinkPhases.push(Math.random() * 2 * Math.PI);
		}
	})();
	
	// a structure storing all the things that could be affecting the camera position
	var cameraPos = {
		'baseX': 0,
		'baseY': 0,
		'driftX': 0,
		'driftY': 0,
		'driftXMag': 0,
		'driftYMag': 0
	};
	window.cameraPos = cameraPos;
	
	function updateCameraPos(t)
	{
		t = t / 5;
		t = t % (2 * Math.PI);
		cameraPos.driftX = cameraPos.driftXMag * Math.cos(t);
		cameraPos.driftY = cameraPos.driftYMag * Math.sin(t * 3);
	}
	window.updateCameraPos = updateCameraPos;
	
	function getTotalCameraPos()
	{
		// add up all the effects on the camera position
		var x = cameraPos.baseX;
		var y = cameraPos.baseY;
		
		x += cameraPos.driftX;
		y += cameraPos.driftY;
		
		return [x, y];
	}
	window.getTotalCameraPos = getTotalCameraPos;
	
	function drawingMeasures(canv)
	{
		// from the width and height of a window, compute the
		// various drawing measurements we need
		
		var result = {};
		var w = canv.width;
		var h = canv.height;
		var scale = Math.min(w / 11, h / 11);
		
		result.w = w;
		result.h = h;
		result.cx = w / 2;
		result.cy = h / 2;
		result.scale = scale;
		return result;
	}
	window.drawingMeasures = drawingMeasures;
	
	function drawGridLayer(canv, ctx, mes, cpos, scaleFactor, gridColor, lineWidth)
	{
		var spacing = scaleFactor * mes.scale;
		var xCount = mes.w / spacing + 2;
		var yCount = mes.h / spacing + 2;
		
		ctx.strokeStyle = gridColor;
		ctx.lineWidth = lineWidth;
		ctx.beginPath();
		for(var i = -Math.ceil(xCount/2); i <= Math.ceil(xCount/2); i++)
		{
			var x = mes.cx + i * spacing
			x -= spacing * cpos[0];
			ctx.moveTo(x, 0);
			ctx.lineTo(x, mes.h);
		}
		for(var i = -Math.ceil(yCount/2); i <= Math.ceil(yCount/2); i++)
		{
			var y = mes.cy + i * spacing
			y -= spacing * cpos[1];
			ctx.moveTo(0, y);
			ctx.lineTo(mes.w, y);
		}
		ctx.stroke();
	}
	
	function drawGrid(canv, t)
	{
		var ctx = canv.getContext('2d');
		var mes = drawingMeasures(canv);
		
		var scaleFactors = [1, 1/2, 1/3, 1/4];
		var gridColors = ["#d8d8d8", "#f0f0f0", "#f4f4f4", "#f8f8f8"];
		var lineWidths = [3/1, 3/2, 3/3, 3/4];
		
		var cpos = getTotalCameraPos();
		
		// draw background layers
		for(var d = scaleFactors.length - 1; d >= 1; d--)
		{
			drawGridLayer(canv, ctx, mes, cpos, scaleFactors[d], gridColors[d], lineWidths[d]);
		}
		
		// draw blinking boxes
		var xCount = mes.w / mes.scale + 4;
		var yCount = mes.h / mes.scale + 4;
		ctx.fillStyle = "#000000"
		for(var i = -Math.ceil(xCount/2); i <= Math.ceil(xCount/2); i++)
		{
			for(var j = -Math.ceil(yCount/2); j <= Math.ceil(yCount/2); j++)
			{
				var idx = (i * 40 + j);
				idx -= BLINK_MODULUS * Math.floor(idx / BLINK_MODULUS);
				var x = mes.cx + i * mes.scale - mes.scale * cpos[0];
				var y = mes.cy + j * mes.scale - mes.scale * cpos[1];
				var phi = t * blinkSpeeds[idx] + blinkPhases[idx];
				var a = 0.05 + Math.cos(phi) * 0.04;
				ctx.globalAlpha = a;
				
				ctx.fillRect(x, y, mes.scale, mes.scale);
			}
		}
		ctx.globalAlpha = 1;
		// draw the front layer
		drawGridLayer(canv, ctx, mes, cpos, scaleFactors[0], gridColors[0], lineWidths[0]);
	}
	window.drawGrid = drawGrid;
	
	window.drawCentreText = function(canv, text)
	{
		var ctx = canv.getContext('2d');
		var w = canv.width;
		var h = canv.height;
		ctx.font = "20pt Helvetica";
		ctx.fillStyle = '#202020';
		ctx.textAlign = 'center';
		ctx.textBaseline = "middle"; 
		ctx.fillText(text, w/2, h/2);
	}
	
	function drawDot(ctx, cx, cy, r)
	{
		ctx.beginPath();
		ctx.arc(cx, cy, r, 0, 2 * Math.PI, false);
		ctx.fill();
	}
    
    window.drawLevelMenu = function(canv, t)
    {
		var ctx = canv.getContext('2d');
		var cpos = getTotalCameraPos();
		var mes = drawingMeasures(canv);
		var alpha = Math.min(1, t - timeModeStarted, timeModeWillEnd - t);
		ctx.globalAlpha = Math.max(alpha, 0);
        
        var nLevs = storyParams.length;
        var cpos = getTotalCameraPos();
        var drawOffX = -2;
        var drawOffY = -1;
        
        var textHeight = 0.45 * mes.scale;
        ctx.fillStyle = "#404040";
        ctx.font = textHeight.toString() + "pt Helvetica";
        ctx.textAlign = 'center';
        ctx.textBaseline = "middle"; 
        ctx.fillText("Select Level", mes.cx + 0.5 * mes.scale, mes.cy + (drawOffY - 0.5) * mes.scale);
        
        for(var i = 0; i < nLevs; i++)
        {
            var cellx = (i % 5);
            var celly = Math.floor(i / 5);
            var drawx = mes.cx + (cellx + drawOffX + 0.5 - cpos[0]) * mes.scale;
            var drawy = mes.cy + (celly + drawOffY + 0.5 - cpos[1]) * mes.scale;
            var r = 0.4 * mes.scale;
            if(canSelectLevel(i))
            {
                if(window.mouseOverLevelNumber == i)
                {
                    ctx.fillStyle = "#202020";
                }
                else
                {
                    ctx.fillStyle = "#404040";
                }
                drawDot(ctx, drawx, drawy, r);
                
                ctx.font = textHeight.toString() + "pt Helvetica";
                ctx.fillStyle = '#c0c0c0';
                ctx.textAlign = 'center';
                ctx.textBaseline = "middle"; 
                ctx.fillText((i + 1).toString(), drawx, drawy);
            }
            else
            {
                ctx.strokeStyle = "#404040";
                ctx.beginPath();
                ctx.arc(drawx, drawy, r, 0, 2 * Math.PI, false);
                ctx.stroke();
            }
        }
        
		ctx.globalAlpha = 1;
    }
	
    var isEdgeSel = {};
	window.drawPlayMode = function(canv, gameState, buttons, t)
	{
		var ctx = canv.getContext('2d');
		var cpos = getTotalCameraPos();
		var mes = drawingMeasures(canv);
		var alpha = Math.min(1, t - timeModeStarted, timeModeWillEnd - t);
		ctx.globalAlpha = Math.max(alpha, 0);
		
		// draw the discrepancies
		var ds = gameState.disc;
		var w = ds.length;
		var h = ds[0].length;
		var drawOffX = theGameState.drawOffset[0];
		var drawOffY = theGameState.drawOffset[1];
		var bc = theGameState.boundingCoords;
		var numDrawn = 0;
		var timeSinceStart = t - theGameState.startedTime;
		var minToDraw;
		var timeSinceEnded = null;
		if(theGameState.endedTime === null)
		{
			minToDraw = -1;
		}
		else
		{
			timeSinceEnded = t - theGameState.endedTime;
			minToDraw = Math.floor((timeSinceEnded - 2) * 50);
		}
		var maxToDraw = Math.floor((timeSinceStart - 0.2) * 50);
		for(var j = bc[1]; j <= bc[3]; j++)
		{
			for(var i = bc[0]; i <= bc[2]; i++)
			{
				numDrawn++;
				var drawx = mes.cx + (i + 0.5 + drawOffX - cpos[0]) * mes.scale;
				var drawy = mes.cy + (j + 0.5 + drawOffY - cpos[1]) * mes.scale;
				if(numDrawn < minToDraw || numDrawn > maxToDraw) continue;
				else if(numDrawn == minToDraw || numDrawn == maxToDraw)
				{
					ctx.fillStyle = "#ffffff";
					ctx.fillRect(drawx - mes.scale / 2, drawy - mes.scale / 2,
						mes.scale, mes.scale);
					ctx.fillStyle = "#aaaaaa";
				}
				else
				{
					var ph = t / 4 + (i+j)/8;
					var lStr1 = Math.floor(Math.cos(ph) * 50 + 120).toString();
					var lStr2 = Math.floor(Math.cos(ph + Math.PI * 1/3) * 50 + 120).toString();
					var lStr3 = Math.floor(Math.cos(ph + Math.PI * 2/3) * 50 + 120).toString();
					ctx.fillStyle = "rgb(" + lStr1 + ',' + lStr2 + ',' + lStr3 + ')';
				}
				var val = ds[i][j];
				if(val == 1)
				{
					drawDot(ctx, drawx, drawy, mes.scale * 0.1);
				}
				else if(val == 2)
				{
					var rotSpeed = theGameState.rotSpeed[i][j];
					var trigx = Math.cos(t * rotSpeed) * mes.scale * 0.12;
					var trigy = Math.sin(t * rotSpeed) * mes.scale * 0.12;
					drawDot(ctx, drawx - trigx, drawy + trigy, mes.scale * 0.1);
					drawDot(ctx, drawx + trigx, drawy - trigy, mes.scale * 0.1);
				}
			}
		}
		if(minToDraw > numDrawn)
		{
			// finished clearing - can start a new level
			playModeNextLevel();
		}
		
		// draw the selected edges
        function drawEdge(e, a)
        {
			x1 = mes.cx + (e.x1 + drawOffX - cpos[0]) * mes.scale;
			y1 = mes.cy + (e.y1 + drawOffY - cpos[1]) * mes.scale;
			x2 = mes.cx + (e.x2 + drawOffX - cpos[0]) * mes.scale;
			y2 = mes.cy + (e.y2 + drawOffY - cpos[1]) * mes.scale;
			ctx.lineWidth = 3;
			if(timeSinceEnded != null)
			{
				ctx.lineWidth = 3 + timeSinceEnded * 15;
				a = a*Math.max(0, (1 - timeSinceEnded));
			}
			ctx.strokeStyle = "rgba(96,96,96,"+a+")";
			ctx.beginPath();
			ctx.moveTo(x1, y1);
			ctx.lineTo(x2, y2);
			ctx.stroke();
        }
        
		var eIdx, e, x1, x2, y1, y2, a;
		for(eIdx = 0; eIdx < gameState.allEdges.length; eIdx++)
		{
			e = gameState.allEdges[eIdx];
            drawEdge(e, 0.1)
		}
		for(eIdx = 0; eIdx < gameState.selEdges.length; eIdx++)
		{
			e = gameState.selEdges[eIdx];
            drawEdge(e, 1)
		}
		
		// draw the highlighted edge if applicable
		e = gameState.highlightEdge;
		if(e != null)
		{
			x1 = mes.cx + (e.x1 + drawOffX - cpos[0]) * mes.scale;
			y1 = mes.cy + (e.y1 + drawOffY - cpos[1]) * mes.scale;
			x2 = mes.cx + (e.x2 + drawOffX - cpos[0]) * mes.scale;
			y2 = mes.cy + (e.y2 + drawOffY - cpos[1]) * mes.scale;
            if(isEdgeSel[e])
            {
                a = Math.cos(t * 10) * 0.1 + 0.85;
            }
            else
            {
                a = Math.cos(t * 10) * 0.1 + 0.15;
            }
			ctx.strokeStyle = "rgba(0,0,0," + a + ")";
			ctx.lineWidth = 4;
			ctx.beginPath();
			ctx.moveTo(x1, y1);
			ctx.lineTo(x2, y2);
			ctx.stroke();
		}
		
		// draw the number of loops required
		var shouldDrawNLoops = !theGameState.storyMode;
		shouldDrawNLoops |= (levelPlayerIsUpTo >= storyShowNLoopsLevel);
		var oldAlpha = ctx.globalAlpha;
		if(levelPlayerIsUpTo == storyShowNLoopsLevel &&
			ctx.globalAlpha == 1)
		{
			// fade in the number of loops
			ctx.globalAlpha = Math.max(0, Math.min(
				1, 0.5 * (t - theGameState.startedTime)));
		}
		if(shouldDrawNLoops)
		{
			var nLoops = theGameState.goalNumLoops;
			var nLoopX = mes.cx - cpos[0] * mes.scale;
			var nLoopY = mes.cy - 4.4 * mes.scale - cpos[1] * mes.scale;
			ctx.fillStyle = "#404040";
			drawDot(ctx, nLoopX, nLoopY, 0.6 * mes.scale);
			
			var textHeight = 0.8 * mes.scale;
			ctx.font = textHeight.toString() + "pt Helvetica";
			ctx.fillStyle = '#c0c0c0';
			ctx.textAlign = 'center';
			ctx.textBaseline = "middle"; 
			ctx.fillText(nLoops, nLoopX, nLoopY);
			
			ctx.fillStyle = '#404040';
			textHeight = 0.5 * mes.scale;
			ctx.font = textHeight.toString() + "pt Helvetica";
			ctx.textAlign = 'right';
			ctx.fillText("Use ", nLoopX - 0.6 * mes.scale, nLoopY);
			ctx.textAlign = 'left';
			ctx.fillText(nLoops>1?" loops": " loop", nLoopX + 0.6 * mes.scale, nLoopY);
		}
		ctx.globalAlpha = oldAlpha;
		
		// draw level text
		if(theGameState.storyMode)
		{
			var levelText = "Level ";
			levelText += (levelPlayerIsUpTo + 1).toString();
			levelText += '/' + (storyParams.length).toString();
		}
		else
		{
			levelText = (infLevelsSolved + 1).toString();
			// level out of infinity (unicode! that's the theme, right?)
			levelText += '/' + String.fromCharCode(0x221e);
		}
		ctx.fillStyle = '#404040';
		textHeight = 0.3 * mes.scale;
		ctx.font = textHeight.toString() + "pt Helvetica";
		ctx.textAlign = 'center';
		ctx.fillText(levelText,
			mes.w - 1.4 * mes.scale - cpos[0] * mes.scale, 0.5 * mes.scale);
		
		// draw buttons
		for(var bId = 0; bId < buttons.length; bId++)
		{
			if(!shouldShowButton(buttons[bId])) continue;
			var cx = mes.w - 1.4 * mes.scale - cpos[0] * mes.scale;
			var cy = (bId * 0.8 + 1) * mes.scale - cpos[1] * mes.scale;
			w = 2.5 * mes.scale;
			h = 0.5 * mes.scale;
			drawButton(canv, cx, cy, w, h, buttons[bId][0], buttons[bId][2]);
		}
		
		ctx.globalAlpha = 1;
	}
	
	window.drawSubtitle = function(canv, isPlayMode, theGameState, t)
	{	
		var ctx = canv.getContext('2d');
		var mes = drawingMeasures(canv);
		var textHeight;
		// draw subtitle or instructions
		var message = null;
		var ss = resources.speechSequence;
		var si = resources.speechIndex;
		if(ss !== null && si !== null && si >= 0 && si < ss.length)
		{
			message = ss[si][1];
		}
		else if(isPlayMode && !theGameState.solved)
		{
			if(theGameState.hint !== null)
			{
				message = theGameState.hint;
			}
			else
			{
				message = theGameState.message;
			}
		}
		if(message != null)
		{
			ctx.fillStyle = '#404040';
			ctx.textBaseline = "middle"; 
			textHeight = 0.35 * mes.scale;
			ctx.font = textHeight.toString() + "pt Helvetica";
			ctx.textAlign = 'center';
			var wrappedText = wordWrap(ctx, message, 10 * mes.scale);
			var textY = mes.cy + 5 * mes.scale - 0.6 * wrappedText.length * mes.scale;
			for(var textLine = 0; textLine < wrappedText.length; textLine++)
			{
				ctx.fillText(wrappedText[textLine], mes.cx, textY);
				textY += 0.6 * mes.scale;
			}
		}
	}
	
	function wordWrap(ctx, text, maxw)
	{
		var outList = [];
		var cLine = '';
		var rem = text;
		while(rem != '')
		{
			// get the next word
			var nextSpace = rem.indexOf(' ');
			var nextWord;
			if(nextSpace < 0)
			{
				nextWord = rem;
				rem = '';
			}
			else
			{
				nextWord = rem.slice(0, nextSpace);
				rem = rem.slice(nextSpace + 1);
			}
			// test if the line is OK
			var trialLine = cLine;
			if(trialLine !== '') trialLine += ' ';
			trialLine += nextWord;
			var textmes = ctx.measureText(trialLine);
			if(cLine === '' || textmes.width < maxw)
			{
				cLine = trialLine;
			}
			else
			{
				outList.push(cLine);
				cLine = nextWord;
			}
		}
		if(cLine != '') outList.push(cLine);
		return outList;
	}
	
	window.drawDoneMode = function(canv, t)
	{
		if(resources.speechSequence !== null)
		{
			// come back when the speech is done
			return;
		}
		
		var ctx = canv.getContext('2d');
		var cpos = getTotalCameraPos();
		var mes = drawingMeasures(canv);
		var textHeight;
		var alpha = Math.min(1, t - resources.lastSpeechEndedTime, timeModeWillEnd - t);
		ctx.globalAlpha = Math.max(alpha, 0);
		
		ctx.fillStyle = '#404040';
		textHeight = 1.2 * mes.scale;
		ctx.font = textHeight.toString() + "pt Helvetica";
		ctx.textAlign = 'center';
		ctx.fillText("The End", mes.cx - cpos[0] * mes.scale,
			mes.cy - cpos[1] * mes.scale - 2.5 * mes.scale);
		
		textHeight = 0.5 * mes.scale;
		ctx.font = textHeight.toString() + "pt Helvetica";
		ctx.fillText("Your rank:", mes.cx - cpos[0] * mes.scale,
			mes.cy - cpos[1] * mes.scale - 1 * mes.scale);
		
		textHeight = 0.7 * mes.scale;
		ctx.font = textHeight.toString() + "pt Helvetica";
		ctx.fillText("Spoony Bard", mes.cx - cpos[0] * mes.scale,
			mes.cy - cpos[1] * mes.scale + 0.2 * mes.scale);
		
		textHeight = 0.5 * mes.scale;
		ctx.font = textHeight.toString() + "pt Helvetica";
		ctx.fillText("Click to return to menu.", mes.cx - cpos[0] * mes.scale,
			mes.cy - cpos[1] * mes.scale + 2 * mes.scale);
		ctx.fillText("Give infinite play a go!", mes.cx - cpos[0] * mes.scale,
			mes.cy - cpos[1] * mes.scale + 3 * mes.scale);
		
			
		ctx.globalAlpha = 1;
	}
	
})();



