(function(){
	"use strict";
	
	var BUTTON_COLOR = "#909090";
	var BUTTON_COLOR_HIGH = "#b0b0b0";
	var BUTTON_TEXT_COLOR = "#282828";
	var BUTTON_TEXT_COLOR_HIGH = "#101010";
	var PARALLAX_SCALE = 1.5;
	var BUTTON_WIDTH_SCALE = 6;
	var BUTTON_HEIGHT_SCALE = 1;
	var BUTTON_STEP = 1.3;
	
	
	window.drawButton = function(canv, cx, cy, width, height, text, highlit)
	{
		var ctx = canv.getContext('2d');
		var r = height / 2;
		var lc = cx - width / 2 + r;
		var rc = cx + width / 2 - r;
		var t = cy - height / 2;
		var b = cy + height / 2;
		ctx.fillStyle = highlit? BUTTON_COLOR_HIGH: BUTTON_COLOR;
		
		ctx.beginPath();
		ctx.rect(lc, t, width - 2 * r, height);
		ctx.arc(rc, cy, r, -Math.PI/2, Math.PI/2, false);
		ctx.arc(lc, cy, r, Math.PI/2, -Math.PI/2, false);
		ctx.fill();
		
		var textHeight = 0.5 * height;
		ctx.font = textHeight.toString() + "pt Helvetica";
		ctx.fillStyle = highlit? BUTTON_TEXT_COLOR: BUTTON_TEXT_COLOR_HIGH;
		ctx.textAlign = 'center';
		ctx.textBaseline = "middle"; 
		ctx.fillText(text, cx, cy);
	}
	
	window.isMouseOverButton = function(mouseX, mouseY, cx, cy, width, height)
	{
		var l = cx - width/2;
		var r = cx + width/2;
		var t = cy - height/2;
		var b = cy + height/2;
		return (mouseX > l && mouseX < r && mouseY > t && mouseY < b);
	}
	
	function drawMainMenu(canv, buttonSpecs, t)
	{
		var ctx = canv.getContext('2d');
		var alpha = Math.min(1, t - timeModeStarted, timeModeWillEnd - t);
		ctx.globalAlpha = Math.max(alpha, 0);
		var mes = drawingMeasures(canv);
		
		var title = resources.data['title'];
		var tAspect = title.height / title.width;
		var th = mes.scale * 1.6;
		var tw = th / tAspect;
		var cpos = getTotalCameraPos();
		// parallax to make the menu appear slightly in front of the grid
		var offx = cpos[0] * mes.scale * PARALLAX_SCALE;
		var offy = cpos[1] * mes.scale * PARALLAX_SCALE;
		
		ctx.drawImage(title, mes.cx - offx - tw / 2,
			2.5 * mes.scale - offy - th/2, tw, th);
			
		for(var b in buttonSpecs)
		{
			var x = mes.cx - offx;
			var y = mes.cy - offy + mes.scale * (BUTTON_STEP * buttonSpecs[b][1] - 0.5);
			var w = mes.scale * BUTTON_WIDTH_SCALE;
			var h = mes.scale * BUTTON_HEIGHT_SCALE;
			drawButton(canv, x, y, w, h, buttonSpecs[b][0], buttonSpecs[b][3]);
		}
		ctx.globalAlpha = 1;
	}
	window.drawMainMenu = drawMainMenu;
	
	window.handleMouseMainMenu = function(canv, mouseX, mouseY, isClick, buttonSpecs)
	{
		var cpos = getTotalCameraPos();
		var mes = drawingMeasures(canv);
		var offx = cpos[0] * mes.scale * PARALLAX_SCALE;
		var offy = cpos[1] * mes.scale * PARALLAX_SCALE;
		
		for(var b in buttonSpecs)
		{
			var bx = mes.cx - offx;
			var by = mes.cy - offy + mes.scale * (BUTTON_STEP * buttonSpecs[b][1] - 0.5);
			var w = mes.scale * BUTTON_WIDTH_SCALE;
			var h = mes.scale * BUTTON_HEIGHT_SCALE;
			if(isMouseOverButton(mouseX, mouseY, bx, by, w, h))
			{
				if(isClick)
				{
                    resources.playSound(resources.data['place']);
					// run the callback
					buttonSpecs[b][2]();
				}
				else
				{
					// set highlit
					buttonSpecs[b][3] = true;
				}
			}
			else
			{
				// don't highlight
				buttonSpecs[b][3] = false;
			}
		}
	};
	
})();

