(function()
{
	"use strict";
	
	var knowGameOver;
	var knowLevelOver;
	
	window.storyModePlayLevel = function(levelId)
	{
		theGameState = {};
		theGameState.storyMode = true;
		var param = storyParams[levelId];
		var w, h;
		if(param.hasOwnProperty('disc'))
		{
			// hand-written level
			theGameState.randomLevel = false;
			theGameState.disc = util.grid2d(levelGen.w, levelGen.h, 0);
			w = param.disc.length;
			h = param.disc[0].length;
			if(typeof h === "undefined")
			{
				throw new Error("Expected a 2d array");
			}
			for(var i = 0; i < w; i++)
			{
				for(var j = 0; j < h; j++)
				{
					theGameState.disc[i][j] = param.disc[i][j];
				}
			}
		}
		else
		{
			// random level
			theGameState.randomLevel = true;
			w = param.w;
			h = param.h;
			theGameState.disc = levelGen.genHardDiscs(
				param.goalNumLoops, w, h, param.thickness);
		}
		theGameState.goalNumLoops = param.goalNumLoops;
		theGameState.drawOffset = [-Math.ceil(w/2), 1-Math.ceil(h/2)];
		if(param.hasOwnProperty('message'))
		{
			theGameState.message = param.message;
		}
		else
		{
			theGameState.message = null;
		}
		
		if(param.hasOwnProperty('speech'))
		{
			resources.playSpeechSequence(param.speech);
		}
		else if(param.hasOwnProperty('speechRepeat'))
		{
			resources.playRepeatingSpeechSequence(param.speechRepeat);
		}
		
		knowGameOver = false;
		commonLevelInit();
        callCmgFunction("start", (levelId + 1).toString());
	}
	
	function commonLevelInit()
	{
		knowLevelOver = false;
		theGameState.allEdges = getEdgeSet(theGameState.disc);
		theGameState.selEdges = [];
		setRotSpeed();
		theGameState.highlightEdge = null;
		theGameState.endedTime = null;
		theGameState.startedTime = timeSecs();
		theGameState.boundingCoords = boundingCoords(theGameState.disc);
		theGameState.solved = false;
		theGameState.hint = null;
	}
	
	function setRotSpeed()
	{
		var w = theGameState.disc.length;
		var h = theGameState.disc[0].length;
		var rotSpeed = util.grid2d(w, h);
		for(var i = 0; i < w; i++)
		{
			for(var j = 0; j < h; j++)
			{
				rotSpeed[i][j] = Math.random() * 0.2 + 0.1;
			}
		}
		theGameState.rotSpeed = rotSpeed;
	}
	
	window.infiniteModeGenerate = function()
	{
		theGameState = {};
		theGameState.randomLevel = true;
		theGameState.storyMode = false;
		var goalNumLoops, genLevelSize, genLevelSize, maxMaxExtends;
		if(Math.random() < 0.5)
		{
			goalNumLoops = Math.floor(Math.random() * 4) + 1;
		}
		else
		{
			goalNumLoops = Math.floor(Math.random() * 5) + 2;
		}
		var minSize = [4, 4, 4, 5, 6, 7, 8][goalNumLoops];
		
		var genLevelSize = Math.floor(Math.random() * (8 - minSize)) + minSize;
		var maxMaxExtends = 20 * genLevelSize + Math.floor(Math.random() * 50);
		
		theGameState.disc = levelGen.genHardDiscs(goalNumLoops,
			genLevelSize, genLevelSize, maxMaxExtends);
		theGameState.goalNumLoops = goalNumLoops;
		var q = Math.ceil(genLevelSize / 2);
		theGameState.drawOffset = [-q, 1-q];
		commonLevelInit();
	}
	
	window.playModeNextLevel = function()
	{
		if(theGameState.storyMode)
		{
			if(theGameState.solved && !knowLevelOver)
			{
				levelPlayerIsUpTo++;
                if(levelPlayerIsUpTo > maxLevelReached)
                {
                    maxLevelReached = levelPlayerIsUpTo;
                }
				storeLevel();
				knowLevelOver = true;
			}
			if(levelPlayerIsUpTo >= storyParams.length)
			{
				if(!knowGameOver)
				{
					gameOver();
					knowGameOver = true;
				}
			}
			else
			{
				storyModePlayLevel(levelPlayerIsUpTo);
			}
		}
		else
		{
			if(theGameState.solved && !knowLevelOver)
			{
				infLevelsSolved++;
				storeLevel();
			}
			// generate a new level in infinite mode
			infiniteModeGenerate();
		}
	}
    
    window.shouldShowButton = function(buttonData)
    {
        if(buttonData[3] == BUTTON_RANDOM && !theGameState.randomLevel) return false;
        if(buttonData[3] == BUTTON_STORY && !theGameState.storyMode) return false;
        return true;
    }
	
	window.handleMousePlay = function(canv, mouseX, mouseY, buttons, isClick)
	{
		var cpos = getTotalCameraPos();
		var mes = drawingMeasures(canv);
		
		// try buttons
		for(var bIdx = 0; bIdx < buttons.length; bIdx++)
		{
			if(!shouldShowButton(buttons[bIdx])) continue;
			var bx = mes.w - 1.4 * mes.scale - cpos[0] * mes.scale;
			var by = (bIdx * 0.8 + 1) * mes.scale - cpos[1] * mes.scale;
			var bw = 2.5 * mes.scale;
			var bh = 0.5 * mes.scale;
			if(isMouseOverButton(mouseX, mouseY, bx, by, bw, bh))
			{
				if(isClick)
				{
					// run the callback
					buttons[bIdx][1]();
				}
				else
				{
					// set highlit
					buttons[bIdx][2] = true;
				}
				return;
			}
			else
			{
				// don't highlight
				buttons[bIdx][2] = false;
			}
		}
		// try edges
		mouseX -= mes.cx;
		mouseY -= mes.cy;
		mouseX /= mes.scale;
		mouseY /= mes.scale;
		var offx = cpos[0];
		var offy = cpos[1];
		mouseX += offx - theGameState.drawOffset[0];
		mouseY += offy - theGameState.drawOffset[1];
		var nearEdge = getNearbyEdge(theGameState.allEdges, mouseX, mouseY)
		
		theGameState.highlightEdge = nearEdge;
		
		if(nearEdge !== null && isClick)
		{
			// add or remove the edge
			var se = theGameState.selEdges;
			var idx = se.indexOf(nearEdge);
			if(idx >= 0)
			{
				se.splice(idx, 1);
			}
			else
			{
				se.push(nearEdge);
			}
		}
		
		// check if the game is finished
		var completeData = isLevelCompleteAndWhyNot(
			theGameState.selEdges, theGameState.disc, theGameState.goalNumLoops);
		var complete = completeData[0];
		theGameState.hint = completeData[1];
		
		if(complete && theGameState.endedTime === null)
		{
			theGameState.endedTime = timeSecs();
			theGameState.solved = true;
			resources.haltSpeechSequence();
			if(theGameState.storyMode && levelPlayerIsUpTo + 1 === storyParams.length)
			{
				// finished the game - play the final speech
				resources.playSpeechSequence(gameOverSequence);
			}
			else
			{
				// play "discrepancy resolved"
				resources.playSpeechSequence(resolvedSpeechSequence);
			}
		}
	}
	
})();

