(function()
{
	"use strict";
	
	var SCALE = 30;
	var OFFX = 160;
	var OFFY = 10;
	
	var GOAL_NUM_LOOPS = 1;
	var GEN_LEVEL_SIZE = 10;
	var MAX_MAX_EXTENDS = 250;
	
	function drawQS(qs)
	{
		var ctx = gameCanvas.getContext('2d');
		var g = qs.grid;
		var w = g.length;
		var h = g[0].length;
		var offsetX = [[0, 10, 20], [20, 10, 20], [20, 10, 0], [0, 10, 0]];
		var offsetY = [[0, 10, 0], [0, 10, 20], [20, 10, 20], [20, 10, 0]];
		for(var i = 0; i < w; i++)
		{
			for(var j = 0; j < h; j++)
			{
				var left = (1 + i) * scale;
				var top = (1 + j) * scale;
				for(var k = 0; k < 4; k++)
				{
					var val = g[i][j][k];
					ctx.beginPath();
					ctx.moveTo(left + offsetX[k][0], top + offsetY[k][0]);
					ctx.lineTo(left + offsetX[k][1], top + offsetY[k][1]);
					ctx.lineTo(left + offsetX[k][2], top + offsetY[k][2]);
					ctx.lineTo(left + offsetX[k][0], top + offsetY[k][0]);
					if(val) ctx.fill();
					ctx.stroke();
				}
			}
		}
	}
	
	function drawDot(ctx, cx, cy, r)
	{
		ctx.beginPath();
		ctx.arc(cx, cy, r, 0, 2 * Math.PI, false);
		ctx.fill();
	}
	
	function drawDiscs(ds)
	{
		var ctx = gameCanvas.getContext('2d');
		ctx.lineWidth = 1;
		var w = ds.length;
		var h = ds[0].length;
		// draw discrepancy dots
		for(var i = 0; i < w; i++)
		{
			for(var j = 0; j < h; j++)
			{
				var left = i * SCALE + OFFX;
				var top = j * SCALE + OFFY;
				ctx.strokeStyle = "#cccccc";
				ctx.fillStyle = "#666666";
				ctx.beginPath();
				ctx.moveTo(left, top);
				ctx.lineTo(left + SCALE, top);
				ctx.lineTo(left + SCALE, top + SCALE);
				ctx.lineTo(left, top + SCALE);
				ctx.lineTo(left, top);
				ctx.stroke();
				if(ds[i][j] == 2)
				{
					drawDot(ctx, left + SCALE/2-7, top + SCALE/2, 5);
					drawDot(ctx, left + SCALE/2+7, top + SCALE/2, 5);
				}
				else if(ds[i][j] == 1)
				{
					ctx.strokeStyle = "#dddddd";
					ctx.beginPath();
					ctx.moveTo(left, top);
					ctx.lineTo(left + SCALE, top + SCALE);
					ctx.moveTo(left + SCALE, top);
					ctx.lineTo(left, top + SCALE);
					ctx.stroke();
					ctx.strokeStyle = "#cccccc";
					drawDot(ctx, left + SCALE/2, top + SCALE/2, 5);
				}
			}
		}
	}
	
	function drawEdges(es, style, lineWidth)
	{
		var ctx = gameCanvas.getContext('2d');
		ctx.lineWidth = lineWidth;
		// draw selected edges
		for(var k = 0; k < es.length; k++)
		{
			var x1 = es[k].x1 * SCALE + OFFX;
			var y1 = es[k].y1 * SCALE + OFFY;
			var x2 = es[k].x2 * SCALE + OFFX;
			var y2 = es[k].y2 * SCALE + OFFY;
			ctx.strokeStyle = style;
			ctx.beginPath();
			ctx.moveTo(x1, y1);
			ctx.lineTo(x2, y2);
			ctx.stroke();
		}
	}

	window.log = function(text)
	{
		var tn = document.createTextNode(text);
		divLog.appendChild(tn);
		var br = document.createElement("br");
		divLog.appendChild(br);
	}
	
	function getGameCoordsOfMouseEvent(evt)
	{
		// get mouse coordinates in canvas coords
        var rect = gameCanvas.getBoundingClientRect();
        var x = evt.clientX - rect.left;
        var y = evt.clientY - rect.top;
        
		// convert to game coords
		x = (x - OFFX) / SCALE;
		y = (y - OFFY) / SCALE;
		
		return [x, y];
	}
	
	function handleMouseMove(evt)
	{
		var mCoords = getGameCoordsOfMouseEvent(evt);
		var edge = getNearbyEdge(allEdges, mCoords[0], mCoords[1]);
		if(edge != null)
		{
			highlightEdge = edge;
		}
		else
		{
			highlightEdge = null;
		}
	}
	
	function handleClick(evt)
	{
		var mCoords = getGameCoordsOfMouseEvent(evt);
		var edge = getNearbyEdge(allEdges, mCoords[0], mCoords[1]);
		if(edge != null)
		{
			var foundIdx = selEdges.indexOf(edge);
			if(foundIdx >= 0)
			{
				selEdges.splice(foundIdx, 1);
			}
			else
			{
				selEdges.push(edge);
			}
		}
		var complete = isLevelComplete(selEdges, discs, GOAL_NUM_LOOPS);
		if(complete)
		{
			log("Success!");
		}
	}
	
	var discs, allEdges, selEdges, highlightEdge;
	
	function frame()
	{
		gameCanvas.width = window.innerWidth / 2;
		gameCanvas.height = window.innerHeight / 2;
		var ctx = gameCanvas.getContext('2d');
		ctx.clearRect(0, 0, gameCanvas.width, gameCanvas.height);
		drawDiscs(discs);
		drawEdges(selEdges, "#606060", 2);
		if(highlightEdge != null)
		{
			if(selEdges.indexOf(highlightEdge) >= 0)
			{
				drawEdges([highlightEdge], "#000000", 2);
			}
			else
			{
				drawEdges([highlightEdge], "#404040", 2);
			}
		}
		
		requestAnimationFrame(frame);
	}
	
	window.onload = function()
	{
		discs = levelGen.genHardDiscs(GOAL_NUM_LOOPS, GEN_LEVEL_SIZE, GEN_LEVEL_SIZE, MAX_MAX_EXTENDS);
		allEdges = getEdgeSet(discs);
		selEdges = [];
		highlightEdge = null;
		
		gameCanvas.addEventListener("mousemove", handleMouseMove);
		gameCanvas.addEventListener("click", handleClick);
		
		requestAnimationFrame(frame);
	}
})();

